@@ -12,12 +12,14 @@ qsplitter  - dummy handler to setup queue tables. All events are ignored. Table
 
 """
 
-from typing import Sequence, Tuple
+from typing import Sequence, Tuple, List, Dict, Optional, Type, Any
 
 from skytools.basetypes import Cursor
 
 import pgq
-from londiste.handler import BaseHandler
+from pgq.baseconsumer import BatchInfo
+from pgq.event import Event
+from londiste.handler import BaseHandler, ApplyFunc
 
 __all__ = ['QueueTableHandler', 'QueueSplitterHandler']
 
@@ -30,13 +32,13 @@ class QueueTableHandler(BaseHandler):
     """
     handler_name = 'qtable'
 
-    def add(self, trigger_arg_list):
+    def add(self, trigger_arg_list: List[str]) -> None:
         """Create SKIP and BEFORE INSERT trigger"""
         trigger_arg_list.append('tgflags=BI')
         trigger_arg_list.append('SKIP')
         trigger_arg_list.append('expect_sync')
 
-    def real_copy(self, tablename, src_curs, dst_curs, column_list):
+    def real_copy(self, tablename: str, src_curs: Cursor, dst_curs: Cursor, column_list: List[str]) -> Tuple[int, int]:
         """Force copy not to start"""
         return (0, 0)
 
@@ -52,7 +54,7 @@ class QueueTableHandler(BaseHandler):
     ) -> Tuple[int, int]:
         return (0, 0)
 
-    def needs_table(self):
+    def needs_table(self) -> bool:
         return False
 
 
@@ -63,8 +65,9 @@ class QueueSplitterHandler(BaseHandler):
       queue=QUEUE - Queue name.
     """
     handler_name = 'qsplitter'
+    rows: List[Sequence[Any]]
 
-    def __init__(self, table_name, args, dest_table):
+    def __init__(self, table_name: str, args: Dict[str, str], dest_table: Optional[str]) -> None:
         """Init per-batch table data cache."""
         super().__init__(table_name, args, dest_table)
         try:
@@ -73,14 +76,14 @@ class QueueSplitterHandler(BaseHandler):
             raise Exception('specify queue with handler-arg') from None
         self.rows = []
 
-    def add(self, trigger_arg_list):
+    def add(self, trigger_arg_list: List[str]) -> None:
         trigger_arg_list.append('virtual_table')
 
-    def prepare_batch(self, batch_info, dst_curs):
+    def prepare_batch(self, batch_info: Optional[BatchInfo], dst_curs: Cursor) -> None:
         """Called on first event for this table in current batch."""
         self.rows = []
 
-    def process_event(self, ev, sql_queue_func, arg):
+    def process_event(self, ev: Event, sql_queue_func: ApplyFunc, dst_curs: Cursor) -> None:
         """Process a event.
 
         Event should be added to sql_queue or executed directly.
@@ -92,7 +95,7 @@ class QueueSplitterHandler(BaseHandler):
                 ev.extra1, ev.extra2, ev.extra3, ev.extra4, ev.time]
         self.rows.append(data)
 
-    def finish_batch(self, batch_info, dst_curs):
+    def finish_batch(self, batch_info: BatchInfo, dst_curs: Cursor) -> None:
         """Called when batch finishes."""
         if self.dst_queue_name is None:
             return
@@ -101,9 +104,9 @@ class QueueSplitterHandler(BaseHandler):
                   'extra1', 'extra2', 'extra3', 'extra4', 'time']
         pgq.bulk_insert_events(dst_curs, self.rows, fields, self.dst_queue_name)
 
-    def needs_table(self):
+    def needs_table(self) -> bool:
         return False
 
 
-__londiste_handlers__ = [QueueTableHandler, QueueSplitterHandler]
+__londiste_handlers__: List[Type[BaseHandler]] = [QueueTableHandler, QueueSplitterHandler]
 
