@@ -3,6 +3,8 @@
 Hack to get local=t for a table, but without processing any events.
 """
 
+from typing import List, Type
+
 from londiste.handler import BaseHandler
 
 __all__ = ['VirtualTableHandler', 'FakeLocalHandler']
@@ -12,10 +14,10 @@ class VirtualTableHandler(BaseHandler):
     __doc__ = __doc__
     handler_name = 'vtable'
 
-    def add(self, trigger_arg_list):
+    def add(self, trigger_arg_list: List[str]) -> None:
         trigger_arg_list.append('virtual_table')
 
-    def needs_table(self):
+    def needs_table(self) -> bool:
         return False
 
 
@@ -24,5 +26,5 @@ class FakeLocalHandler(VirtualTableHandler):
     handler_name = 'fake_local'
 
 
-__londiste_handlers__ = [VirtualTableHandler, FakeLocalHandler]
+__londiste_handlers__: List[Type[BaseHandler]] = [VirtualTableHandler, FakeLocalHandler]
 
