@@ -2,7 +2,7 @@
 
 import londiste
 
-from nose.tools import *
 
 def test_one():
-    assert_true(True)
+    assert londiste
+
